/*-------------------------------------------------------------------------
--
--  checker_pec.c - The entry point for the runtime DNS pattern analysis
--  system.  This file also includes some functions to aid decoding
--  of the DNS message.
--
--  Copyright (c) 1992, 1993, 1994, 1995
--                University of Southern California, Los Angeles, CA
--                All rights reserved.
--
--  Redistribution and use of this software in source or binary form is
--  permitted given that the above copyright notice and this paragraph are
--  duplicated in all such forms and that any documentation and material
--  related to such distribution and use acknowledge that the software
--  was developed by University of Southern California.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMIT TO, THE IMPLIED
--  WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--  THE UNIVERSITY AND THE AUTHORS ASSUME NO RESPONSIBILITY FOR DAMAGES
--  OF ANY KIND RESULTING FROM THE USE OF THIS SOFTWARE.
--
--  Please contact Dr. Peter Danzig of Computer Science Department at
--  danzig@usc.edu for further information.
--
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   | 12/5/92  | Initial version                | Steve Miller - USC --
--  1   |  2/94    | added PEC, separated to 2 prcs | Charles Chan - USC --
--  2   |  5/95    | Minor debug, new log/config    | Steve Miller - USC --
-------------------------------------------------------------------------*/

#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <varargs.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include "checker.h"


/*-------------------------------------------------------------------------
--
-- variable declarations
--
-------------------------------------------------------------------------*/

static char     srvname[200] = "\0";
static u_long   srvaddr = 0;
static u_long   checker_enabled = 0;
char            base_password[50];
u_long          negative_cache_time;
u_long          password_check;
u_long          local_domain_checking = 0;
u_long          pec_etc_hosts = 0;
u_long          pec_checking;
u_long          pecImmunePeriod;
u_long          pecTestPeriod;
int             pecIntervalNum;
int             rcodeErrorChecking = 0;
int             packetCount = 0;

extern int      errno;
extern char    *sys_errlist[];
time_t          current_time;
PEC_LIST_TYPE  *pecList;
PEC_LIST_TYPE  *pecImmuneList;

char   pecResultFileName[100] = DEFAULT_PEC_RESULT;
char   pecStatusFileName[100] = DEFAULT_PEC_STATUS;
char   pecLogFileName[100]    = DEFAULT_PEC_LOG;

FILE  *pecFile;    /* for file to log pec status */
FILE  *logFile;    /* for log file */

#define THIS_NET_CLASS  IN_CLASSC_NET
#define MAX_SERVER_NUMBER 5

struct _serverList {
    char            server_name[20];
    int             sockfd;
    int             error_count;
    } serverList[MAX_SERVER_NUMBER];

char *qtypeString[] = {
    " [UNKNOWN]", " [A]",      " [NS]",    " [MD]",
    " [MF]",      " [CNAME]",  " [SOA]",   " [MB]",
    " [MG]",      " [MR]",     " [NULL]",  " [WKS]",
    " [PTR]",     " [HINFO]",  " [MINFO]", " [MX]",
    " [TXT]",     " [AXFR]",   " [MAILB]", " [MAILA]",
    " [*]" };

u_long          get_answer_ttl();
char           *get_query();
u_short         q_type();
u_long          checkergethostbyname();
void            LogMessage();
void            GetServer();

static char aeqstr[] =
"===============================================================";


/*------------------------------------------------------------------------
--
--  checker() - The entry point into the set of functions that
--  update and maintain the query/response database.
--
--  parameters:
--      ctype - flag showing query/response (query=0, response=1)
--      mptr - pointer to the message to be processed.
--      mlen - length in bytes of message
--      saddr - socket address structure.
--      proto - flag showing protocol type.
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
u_long
checker(ctype, mptr, mlen, saddr, qsp)
u_long ctype;
struct qstream *qsp;
struct sockaddr_in *saddr;
char  *mptr;
int    mlen;
    {
    QUERY_PACKET    qp;        /* record for the query */
    SOURCE_INFO     src;       /* record for the source of the query */
    struct timeval  tm;        /* time record */
    HEADER         *dh;        /* dns message header */
    char           *ptr,       /* buffer pointers */
                   *cp,
                   *eom;
    int             remain,
                    control_code;
    static int      init = 0;
    u_short         typr,
                    qclass;
    char            name[1024],
                    ts[60],
                    tnam[30];
    u_long          ptype;
    struct hostent *hptr = NULL;

    char            server[20];
    u_short         port = 0;

    int             pecReturnCode;  /* PecControl() return */
    int             i;              /* index */
    struct tm      *tm_time;
    char            rcodeFlag;

    /* If we haven't initialized yet then do so. */

    if (!init) {

        /*
         * Open up a log file, which is left open so as not to
         * slow down the nameserver too much with logging calls.
         */

        if ((logFile = fopen(pecLogFileName, "a")) == NULL)
            fprintf(stderr, "checker-pec cannot open log file\n");

        /*
         * Log the fact that we have started up.
         */

        LogMessage(aeqstr);
        LogMessage("checker started");

        /*
         * Do an initial read of the configuration file from
         * the /etc/checker.cfg or from a local checker.cfg file.
         */

        pec_read_configuration();

        /* Get and log our own host name. */

        gethostname(srvname, sizeof(srvname));
        LogMessage("gethostname = %s", srvname);

        /*
         * If the user hasn't disabled it then do a name translation
         * on our name.
         */

        if( ! pec_etc_hosts ){
            hptr = gethostbyname(srvname);
            if( hptr == NULL )
                 LogMessage("gethostbyname: %s", sys_errlist[errno]);
            else LogMessage("gethostbyname is okay");
            }

        /*
         * Get our host's address, either from the name translation
         * address structure or out of the /etc/hosts file.
         */

        if (hptr != NULL) srvaddr = *(u_long *) hptr->h_addr_list[0];
        else              srvaddr = checkergethostbyname(srvname);

        /* Make sure our PEC lists are initialized and set our flag */

        init = 1;
        pecList = NULL;
        pecImmuneList = NULL;
        for (i = 0; i < MAX_SERVER_NUMBER; i++) {
            serverList[i].server_name[0] = '\0';
            serverList[i].sockfd = -1;
            }
        }

    /* Make sure we don't have garbage in our control structures. */

    memset(&qp, 0, sizeof(QUERY_PACKET));
    memset(&src, 0, sizeof(SOURCE_INFO));

    /* Set up pointers to the DNS message header and body */

    dh = (HEADER *) mptr;
    ptr = mptr + sizeof(HEADER);
    cp = ptr;
    eom = mptr + mlen;
    remain = mlen - sizeof(HEADER);

    /* Fill in the current time, and host address. */

    gettimeofday(&tm, (struct timezone *) 0);
    src.current_time = tm.tv_sec;
    src.our_address = srvaddr;
    current_time = src.current_time;

    /* Get the remote node's IP address and port number. */

    src.ip_address = saddr->sin_addr.s_addr;
    src.portnum = ntohs(saddr->sin_port);

    /*
     * Extract the query name and type from the message, if we get an
     * error here then assume the message is badly formed and just 
     * ignore it.
     */

    if (get_query(ptr, remain, &ptype, &qclass, name) == NULL)
        return(0);

    /* Check the returned name for nul */

    if (strcmp(name, "") == 0) { strcpy(name, "NONE"); }

    /*
     * Look to see if the query name is in the form of a checker control
     * command.  If so then make sure this is the query and not the 
     * response packet (no need to execute it twice).  After that look
     * to see if we are supposed to find a password on the back of the 
     * command string.  Then, if everything else is ok, execute the 
     * command.
     */

    if (strncmp(name, "checker.control.command.", 24) == 0) {

        if (ctype == 0)
            {
            if (password_check && !validate_password(name));
            else
                {
                sscanf(&name[24], "%d", &control_code);
                switch (control_code)
                    {

                    /*
                     * checker.control.command.1
                     * have the local code and the backend database processors
                     * reread their configuration files
                     */

                    case 1:
                        pec_read_configuration();
                        break;

                    /*
                     * checker.control.command.2
                     * have the local PEC code print a report
                     */

                    case 2:
                        if ((pecFile = fopen(pecStatusFileName, "a")) == NULL) {
                            LogMessage("checker cannot open PEC report file.");
                            return(0); }
                        PecPrintList(pecList, pecFile);
                        PecPrintList(pecImmuneList, pecFile);
                        fclose(pecFile);
                        break;

                    /*
                     * checker.control.command.3
                     * open a connection to a remote checker database processor
                     */

                    case 3:
                        i = NextServerSlot();
                        if (i >= 0) {
                            server[0] = '\0';
                            GetServer(&name[25], server, &port);
                            strcpy(serverList[i].server_name, server);
                            OpenConnection(&serverList[i].sockfd,
                                serverList[i].server_name, port);
                            serverList[i].error_count = 0;
                            }
                        else
                            LogMessage("Cannot open more servers, no space!");
                        break;

                    /*
                     * checker.control.command.4
                     * close down a connection to a remote checker database
                     */

                    case 4:
                        CloseConnection(&name[26]);
                        break;

                    /*
                     * checker.control.command.5
                     * cause the remote database processors to print a report
                     */

                    case 5:
                        break;

                    /*
                     * checker.control.command.???
                     */

                    default:
                        break;
                    }

                /*
                 * If we passed the security check then always send the control
                 * packet on to the backend database processors to do with it
                 * what they want.  This prevents this program from having to
                 * know too much about the backend code.
                 */

                qp.type = ctype;
                strcpy(qp.qname, name);
                qp.qtype = q_type(&ptype);
                SendPacket(&src, &qp);
                LogMessage("checker control command passed on");
                }
            }
        return(0);
        }

    /*
     * If our enable/disable flag is not set then return to the caller with
     * no further actions.
     */

    if( !checker_enabled ) return(0);

    /* Move the query type and name into the structure fields. */

    qp.type = ctype;
    strcpy(qp.qname, name);
    qp.qtype = q_type(&ptype);

    /* adjust to valid range of ptype */

    if (qp.qtype > 20) qp.qtype = 0;

    /* Get the fields in the header of the DNS message packet. */

    qp.id = dh->id;
    qp.aa = dh->aa;
    qp.rcode = dh->rcode;
    qp.qdcount = dh->qdcount;
    qp.ancount = dh->ancount;
    qp.nscount = dh->nscount;
    qp.arcount = dh->arcount;

    /* get a value to use in some of the cache and error checks. */

    if (dh->rcode == NOERROR)
         qp.ttl = get_answer_ttl(ptr, eom, &cp, (int) dh->ancount);
    else qp.ttl = negative_cache_time;

    /*
     * Figure out what protocol we are dealing with and if the message is a
     * response then parse out the nameserver fields to see if our own name
     * is in there.
     */

    if (qsp == (struct qstream *) 0) src.protocol = 'U';
    else                             src.protocol = 'T';
    if (ctype == RESPONSE_MESSAGE)
        src.ourname = get_ourname(ptr, eom, &cp, (int) dh->nscount);
    pecReturnCode = 0;

    if (pec_checking) {

        rcodeFlag = 1;
        if (rcodeErrorChecking && qp.rcode != 0) rcodeFlag = 0;

        /*
         * The status is an indication of whether or not the query name is
         * under interdiction by the PEC algorithm.  Return it to the caller
         */

        if (ctype == 1 && rcodeFlag) {

            /* drop it if we don't want to check the local domain */

            if( !(local_domain_checking && 
                 (THIS_NET_CLASS & src.our_address) == 
                 (THIS_NET_CLASS & src.ip_address))) {

                /* append the type name to the end of the query name */

                strcat(qp.qname, qtypeString[qp.qtype]);
                pecReturnCode = PecControl(src.ip_address, qp.qname);
                }
            }

        if (pecReturnCode > 0) {
            sprintf(name, "** PEC captured [%d]: ", pecReturnCode);
            strcat(name, qp.qname);
            strcpy(qp.qname, name);
            pecReturnCode = 1;
            }
        }

    SendPacket(&src, &qp);
    return(pecReturnCode);
    }


/*------------------------------------------------------------------------
--
-- int NextServerSlot()
--     returns an empty slot in the server list for the client/server
--     communication for the query info
--
--
--  GIVEN          : none
--  GLOBAL USED    : serverList
--  RETURNS        : an empty slot number in serverList if there is one,
--                   -1 if not
--  POST CONDITIONS: none
--
--  COMMENTS:
--     none
--
-------------------------------------------------------------------------*/
int
NextServerSlot()
    {
    int i;

    for (i = 0; i < MAX_SERVER_NUMBER; i++)
        if (serverList[i].sockfd < 0) return i;

    return -1;
    }


/*-------------------------------------------------------------------------
--
-- u_long checkergethostbyname()
--     parses /etc/hosts file to get the ip address given by name;
--     it returns the ip address in the network ordered hex address.  Note
--     that we may not want to use gethostbyname() because, when we start
--     the named program and NIS is running, gethostbyname() may return NULL
--
--  GIVEN          : dns name
--  GLOBAL USED    : none
--  RETURNS        : ip address in hex format, 0 if no matching name
--
--  POST CONDITIONS: none
--
--  COMMENTS:
--     use /etc/hosts file
--
---------------------------------------------------------------------------*/
u_long
checkergethostbyname(name)
char *name;
    {
    int i;
    u_long addr;
    FILE *fp;
    char  lstr[100];
    char  line[250];
    char  da[100];
    char  dn[10][100];

    if ((fp = fopen("/etc/hosts", "r")) != NULL)
        {

        while (fgets(line, 250, fp) != NULL)
            {
            memset( dn, 0, sizeof(dn) );
            sscanf( line, "%s %s %s %s %s %s %s %s %s %s %s", da, 
                    dn[0], dn[1], dn[2], dn[3], dn[4],
                    dn[5], dn[6], dn[7], dn[8], dn[9] );

            for( i=0; i<10; i++ )
                if( strcmp( name, dn[i] ) == 0 ) break;
            if( i == 10 ) continue;

            addr = inet_addr(da);
            sprintf( lstr, "checkergethostbyname found %s", da );
            LogMessage( lstr );
            fclose(fp);
            return( addr );
            }
        
        sprintf( lstr, "checkergethostbyname could not find %s", name );
        LogMessage( lstr );
        fclose(fp);
        }

    return 0;
    }


/*-------------------------------------------------------------------------
--
-- void LogMessage()
--     function writes message to logFile.  It talks argument list
--
--  GIVEN          : va_list
--  GLOBAL USED    : none
--  RETURNS        : noe
--
--  POST CONDITIONS: none
--
--  COMMENTS:
--     none
--
---------------------------------------------------------------------------*/
void
LogMessage(va_alist)
va_dcl
    {
    va_list     args;     /* arguments list */
    char       *fmt;      /* pointer format string */
    struct tm  *tm_time;  /* stores time in tm format    */
    time_t      now;      /* time stamp */

    if (logFile == NULL) return;

    now = time((time_t *) 0);
    tm_time = localtime(&now);

    fprintf(logFile, "%02d%02d%02d-%02d:%02d:%02d  ",
        tm_time->tm_year, tm_time->tm_mon + 1, tm_time->tm_mday,
        tm_time->tm_hour, tm_time->tm_min, tm_time->tm_sec);

    va_start(args);
    fmt = va_arg(args, char *);
    vfprintf(logFile, fmt, args);
    fputc('\n', logFile);
    va_end(args);

    fflush(logFile);
    }


/*-------------------------------------------------------------------------
--
-- void GetServer()
--     parses checker.control.command.2-<server>-<port #>
--
--  GIVEN          : string to parse beginning with "-" (-<server>-<port #>)
--  GLOBAL USED    : none
--  RETURNS        : none, but values of server and port will be updated
--
--  POST CONDITIONS: none
--
--  COMMENTS:
--     none
--
---------------------------------------------------------------------------*/
void
GetServer(str, server, port)
char       *str;
char       *server;
u_short    *port;
    {
    char   *ptrStr;
    char   *ptrServer;

    /* assgin the pointers */

    ptrStr = str;
    ptrServer = server;

    /* get rid something we don't want */

    while (!isdigit(*ptrStr)) *ptrStr++;

    /* get the server IP address string */

    while (*ptrStr != '-') *ptrServer++ = *ptrStr++;

    /* terminate server string */

    *ptrServer = '\0';

    /* get the port number */

    *ptrStr++;
    sscanf(ptrStr, "%hd", port);
    }


/*-------------------------------------------------------------------------
--
-- int WriteNBytes()
--     writes nbytes of the given ptr to the given fd.
--
--  GIVEN          : 
--  GLOBAL USED    : 
--  RETURNS        : 
--
--  POST CONDITIONS: none
--
--  COMMENTS:
--     To make sure all the info in the buffer is written to a socket (thus,
--     to guranttee the bytes ordering, the mechanic in this function may be
--     used.
--
---------------------------------------------------------------------------*/
int
WriteNBytes(fd, ptr, nBytes)
int    fd;
char  *ptr;
int    nBytes;
    {
    int nLeft, nWritten;

    nLeft = nBytes;

    while( nLeft > 0 ) {

        nWritten = write(fd, ptr, nLeft);

        if( nWritten <= 0 ) {
           if(nLeft == nBytes) return( nWritten );
           else                return( nBytes - nLeft ); }

        nLeft -= nWritten;
        ptr += nWritten;
        }

    return( nBytes - nLeft );
    }


/*-------------------------------------------------------------------------
--
-- int OpenConnection()
--     opens a tcp connection to a server
--
--  GIVEN          : server address and server port number
--  GLOBAL USED    : none
--  RETURNS        : none, but socketfd will be updated
--
--  POST CONDITIONS: none
--
--  COMMENTS:
--    none
--
---------------------------------------------------------------------------*/
int
OpenConnection(sockfd, serv_address, serv_port)
int     *sockfd;
char    *serv_address;
u_short  serv_port;
    {
    struct sockaddr_in serv_addr;

    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr(serv_address);
    serv_addr.sin_port = htons(serv_port);

    /*
     * Open a TCP socket (an Internet stream socket).
     */

    *sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (*sockfd < 0) return;

    /*
     * Connect to the server.
     */

    connect(*sockfd, (struct sockaddr *) & serv_addr, sizeof(serv_addr));
    LogMessage("connection to server %s into port %d", serv_address, serv_port);
    }


/*-------------------------------------------------------------------------
--
-- int CloseConnection()
--     closes a tcp connection opened by OpenConnection()
--
--  GIVEN          : server to close
--  GLOBAL USED    : none
--  RETURNS        : none
--
--  POST CONDITIONS: the used slot in serverList will be available again
--
--  COMMENTS:
--    none
--
---------------------------------------------------------------------------*/
int
CloseConnection(server)
char  *server;
    {
    int    i;

    /* server = "0" means to close all the connection */

    for (i = 0; i < MAX_SERVER_NUMBER; i++) {

    if (!strcmp(serverList[i].server_name, server) || !strcmp(server, "0"))

        if (serverList[i].sockfd >= 0) {
            close(serverList[i].sockfd);
            serverList[i].sockfd = -1;
            serverList[i].error_count = 0;
            LogMessage("closes connection to %s", serverList[i].server_name);
            }

        }
    }


/*-------------------------------------------------------------------------
--
-- int SendPacket()
--     function sends out src and qp packets to all the servers in serverList
--
--  GIVEN          : src, qp
--  GLOBAL USED    : serverList
--  RETURNS        : none
--
--  POST CONDITIONS: none
--
--  COMMENTS:
--    none
--
---------------------------------------------------------------------------*/
int
SendPacket(src, qp)
SOURCE_INFO    *src;
QUERY_PACKET   *qp;
    {
    int  i;
    int  n;

    for (i = 0; i < MAX_SERVER_NUMBER; i++) {

        if (serverList[i].sockfd >= 0) {
    
            n = sizeof(SOURCE_INFO);

            if (WriteNBytes(serverList[i].sockfd, (char *) src, n) != n) {
                LogMessage("cannot send source packet to %s", serverList[i].server_name);
                serverList[i].error_count++;
                }

            n = sizeof(QUERY_PACKET);

            if (WriteNBytes(serverList[i].sockfd, (char *) qp, n) != n) {
                LogMessage("cannot send query packet to %s", serverList[i].server_name);
                serverList[i].error_count++;
                }

            if (serverList[i].error_count > 100) {
                LogMessage("too many error counts in sending packets to %s", serverList[i].server_name);
                LogMessage("CloseConnection() for %s called", serverList[i].server_name);
                CloseConnection(serverList[i].server_name);
                }
            }
        }
    }


/*-----------------------------------------------------------------------
--
-- int pec_read_configuration()
--     A routine to read the configuration file and set the variables
--     as indicated.
--
--  GIVEN          : 
--  GLOBAL USED    : 
--  RETURNS        : 
--
--  POST CONDITIONS: 
--
--  COMMENTS:
--    none
--
-----------------------------------------------------------------------*/
int
pec_read_configuration()
    {
    FILE      *fp;
    char       line[100];
    char      *t1,
              *t2,
               t3[100];
    u_long     tmp;
    char       tmp1[100];

    /* Go find the configuration file. */

    if ((fp = fopen("/etc/checker.cfg", "r")) == NULL) {
        if ((fp = fopen("./checker.cfg", "r")) == NULL) {
            LogMessage("cannot open checker.cfg");
            return; }
        else LogMessage("reading configuration from ./checker.cfg"); }
    else LogMessage("reading configuration from /etc/checker.cfg");

    /* Get a line and get the third token as a parameter. */

    while (fgets(line, 100, fp) != NULL)
        {
        if (line[0] == '#') continue;
        sscanf(line, "%*s %*s %s", t3);

        if (strncmp(line, "password check", 14) == 0) {
            if (strncmp("off", t3, 3) == 0) password_check = 0;
            else                            password_check = 1;
            LogMessage("set password check flag to %d", password_check);
            }

        else if (strncmp(line, "password string", 15) == 0) {
            sscanf(t3, "%s", base_password);
            LogMessage("encrypted password string is now %s", base_password);
            }

        else if (strncmp(line, "domain checking", 15) == 0) {
            if (strncmp("off", t3, 3) == 0) local_domain_checking = 0;
            else                            local_domain_checking = 1;
            LogMessage("domain checking flag is now %d.", local_domain_checking);
            }

        else if (strncmp(line, "checker enabled", 15) == 0) {
            if (strncmp("off", t3, 3) == 0) checker_enabled = 0;
            else                            checker_enabled = 1;
            LogMessage("checker enable flag is now %d.", checker_enabled);
            }

        else if (strncmp(line, "negative cache", 14) == 0) {
            sscanf(t3, "%d", &negative_cache_time);
            LogMessage("negative_cache_time is %d", negative_cache_time);
            }

        /* read the PEC variables */

        else if (strncmp(line, "pec checking", 12) == 0) {
            if (strncmp("off", t3, 3) == 0) pec_checking = 0;
            else                            pec_checking = 1;
            LogMessage("pec checking enable flag is now %d.", pec_checking);
            }

        else if (strncmp(line, "pec test-period", 15) == 0) {
            sscanf(t3, "%ld", &pecTestPeriod);
            LogMessage("pecTestLength = %ld", pecTestPeriod);
            }

        else if (strncmp(line, "pec immuned-period", 18) == 0) {
            sscanf(t3, "%ld", &pecImmunePeriod);
            LogMessage("pecImmunePeriod = %ld", pecImmunePeriod);
            }

        else if (strncmp(line, "pec result-file", 15) == 0) {
            strcpy(pecResultFileName, t3);
            LogMessage("pecResultFilename = %s", pecResultFileName);
            }

        else if (strncmp(line, "pec status-file", 15) == 0) {
            strcpy(pecStatusFileName, t3);
            LogMessage("pecStatusFilename = %s", pecStatusFileName);
            }

        else if (strncmp(line, "pec use-hosts", 13) == 0) {
            if (strncmp("off", t3, 3) == 0) pec_etc_hosts = 0;
            else                            pec_etc_hosts = 1;
            LogMessage("set pec_etc_hosts flag to %d", pec_etc_hosts );
            }

        else if (strncmp(line, "pec log-file", 12) == 0) {
            if( strcmp( t3, pecLogFileName ) != 0 ) {
                strcpy(pecLogFileName, t3);
                LogMessage("new pec log file name = %s", pecLogFileName);
                fclose( logFile );
                if ((logFile = fopen(pecLogFileName, "a")) == NULL)
                    fprintf(stderr, 
                    "checker-pec cannot open log file %s\n", pecLogFileName );
                }
            else
                LogMessage("pec log file name = %s", pecLogFileName);
            }
    
        else if (strncmp(line, "pec interval", 12) == 0) {
            sscanf(t3, "%d", &pecIntervalNum);
            LogMessage("pecIntervalNum = %d", pecIntervalNum);
            }

        else if (strncmp(line, "pec rcode-checking", 18) == 0) {
            if (strncmp("off", t3, 3) == 0) rcodeErrorChecking = 0;
            else                            rcodeErrorChecking = 1;
            LogMessage("rcode error checking flag is now %d", rcodeErrorChecking);
            }

        }
    fclose(fp);
    }


/*-----------------------------------------------------------------------
--
-- u_long get_answer_ttl()
--     A little routine to pull the time-to-live for the answer record
--     from a returned response.
--
--  GIVEN          : 
--  GLOBAL USED    : 
--  RETURNS        : 
--
--  POST CONDITIONS: 
--
--  COMMENTS:
--    none
--
-----------------------------------------------------------------------*/
u_long
get_answer_ttl(ptr, eom, cp, ancount)
char        *ptr;
char        *eom;
char       **cp;
int          ancount;
    {
    u_long   ttl;
    u_long   tttl;
    u_long   _getlong();
    u_short  _getshort();
    int      rdlen, n;

    /*
     * Assume that there is only one question record and get its fields out
     * of the way.
     */

    if ((n = dn_skipname(ptr, eom)) < 0) return (negative_cache_time);
    *cp += n;                  /* account for name field     */
    *cp += sizeof(u_short);    /* account for type field     */
    *cp += sizeof(u_short);    /* account for class field    */

    /*
     * Now loop for the number of answer records, trying to find the one with
     * the least TTL.  That is the one we want to use for database limiting
     * and cache leak detection.
     */

    for (ttl = 0xEFFFFFFF; ancount > 0; ancount--) {

        /*
         * Read out the name field from the answer resource record. The name
         * itself isn't really needed, but we have to account for the space.
         */

        if ((n = dn_skipname(*cp, eom)) < 0) break;

        /*
         * Now get the ttl value and account for the rest of the resource
         * record.
         */

        *cp += n;                    /* account for name field     */
        *cp += sizeof(u_short);      /* account for type field     */
        *cp += sizeof(u_short);      /* account for class field    */
        if ((*cp) + 3 > eom) break;  /* make sure we are in msg    */
        tttl = _getlong(*cp);        /* get time to live value     */
        *cp += sizeof(u_long);       /* account for ttl field      */
        if ((*cp) + 3 > eom) break;  /* make sure we are in msg    */
        rdlen = _getshort(*cp);      /* get length of rdata field  */
        *cp += sizeof(u_short);      /* account for rdlength field */
        *cp += rdlen;                /* account for rdata fields   */

        /*
         * If this was a real time to live value, and is less than any other
         * we have seen then save it.
         */

        if (tttl > 0 && tttl < ttl) ttl = tttl;
        }

    /* If we didn't find any ttl then return the negative ttl. */

    if (ttl == 0xEFFFFFFF) ttl = negative_cache_time;

    /* Give the ttl value back to the caller */

    return (ttl);
    }


/*-----------------------------------------------------------------------
--
-- int get_ourname();
--     A little routine to see if our own server name is in the authority
--     records section of the response.
--
--  GIVEN          : 
--  GLOBAL USED    : 
--  RETURNS        : 
--
--  POST CONDITIONS: 
--
--  COMMENTS:
--    none
--
-----------------------------------------------------------------------*/
int
get_ourname(ptr, eom, cp, nscount)
char        *ptr;
char        *eom;
char       **cp;
int          nscount;
    {
    u_long   ttl;
    u_long   tttl;
    u_long  _getlong();
    u_short _getshort();
    int      rdlen, n, found, done, type;
    char     nbuf[200];


    /* Now loop for the number of authority records */

    for (found = 0, done = 0; !done && nscount > 0; nscount--) {

        /*
         * Skip the name field from the resource record. The name itself
         * isn't really needed, but we have to account for the space.
         */

        if ((n = dn_skipname(*cp, eom)) < 0) break;

        /* Now account for the rest of the resource record. */

        *cp += n;                 /* account for name field     */
        if (*cp >= eom) break;    /* make sure we don't overrun */
        type = _getshort(*cp);    /* get the record's type      */
        *cp += sizeof(u_short);   /* account for type field     */
        *cp += sizeof(u_short);   /* account for class field    */
        *cp += sizeof(u_long);    /* account for ttl field      */
        if (*cp >= eom) break;    /* make sure we don't overrun */
        rdlen = _getshort(*cp);   /* get length of rdata field  */
        *cp += sizeof(u_short);   /* account for rdlength field */

        /*
         * If this is really an NS type of record then expand the domain name
         * in the rdata field to see if it matches our own name.
         */
    
        if (type == T_NS) {

            dn_expand(ptr, eom, *cp, nbuf, sizeof(nbuf));

            if (strncmp(nbuf, srvname, strlen(srvname)) == 0) {
                done = 1;
                found = 1;
                }
            }

        *cp += rdlen;        /* account for rdata fields   */
        }

    /* Give the found/not-found status back to the caller */

    return (found);
    }


/*------------------------------------------------------------------------
--
--  get_query - a routine to extract the contents of the question type of
--  resource record that is pointed to by the input pointer.
--
--  Note that this routine assumes that this is the first name in the
--  message and does not attempt to discover or resolve compressed
--  labels.
--
--  GIVEN          : 
--  GLOBAL USED    : 
--  RETURNS        : 
--
--  POST CONDITIONS: 
--
--  COMMENTS:
--      anant jan 19, modify to also accept qclass pointer as an
--      argument. class value is written and thus passed back.
--
-------------------------------------------------------------------------*/
char*
get_query(ptr, remain, typep, qclass, name)
char           *ptr;
int             remain;
u_short        *qclass;
u_short        *typep;
char           *name;
    {
    int             i;
    int             qlen, len;
    char            label[256];
    int             cnt = 0;

    /* Make sure we start out clean. */

    *name = '\0';
    *label = '\0';

    /*
     * Loop through all of the labels (those pieces normally separated by
     * dots in our common Domain Name structure.
     */

    for (i = 0;; i++) {

        /* Get the length of the label and check that its not too long */

        qlen = (u_char) (*ptr);
        if (qlen > 63) {
            sprintf(name, "LONG COMPONENT");
            return (name);
            }

        /*
         * Add that to the total length and check that the whole name is not
         * too long.
         */

        cnt += qlen;
        if (cnt > 1024) {
            sprintf(name, "LONG NAME");
            return (name);
            }

        /*
         * Update our byte pointers into the message, and break out of the
         * loop if we have found the end of the name.
         */

        ptr++;
        remain--;
        if (qlen == 0) break;

        /* Copy this component of the name into our working buffer. */

        bcopy(ptr, label, qlen);
        label[qlen] = '\0';
        ptr += qlen;
        remain -= qlen;

        /*
         * If this is the first component of the name then just copy the
         * label into the output buffer.  If not then add the dot before this
         * component and concatenate with the current contents of the output
         * buffer.
         */

        if (i == 0) {
            strncpy(name, label, qlen + 1); }
        else {
            strcat(name, ".");
            strncat(name, label, qlen + 1); }
        }

    /*
     * Now recover the query type and class from the end of the record in the
     * message.
     */

    len = sizeof(u_short);
    bcopy(ptr, typep, len);
    ptr += len;
    remain -= len;
    bcopy(ptr, qclass, len);
    ptr += len;
    remain -= len;
    return (name);
    }


/*-----------------------------------------------------------------------
--
--  q_type - decode the query type field in a question record
--
--  GIVEN          : 
--  GLOBAL USED    : 
--  RETURNS        : 
--
--  POST CONDITIONS: 
--
--  COMMENTS:
--      none
--
-----------------------------------------------------------------------*/
u_short
q_type(typep)
u_short   *typep;
    {
    u_short vtype;

    if ((*typep < 17) && (*typep > 0)) 
         vtype = *typep;

    else if ((*typep > 251) && (*typep < 256))
         vtype = *typep - 235;    /* - 252 + 16 + 1 */

    else vtype = 0;

    return (vtype);
    }


/*-----------------------------------------------------------------------
--
-- int validate_password()
--     A little routine to check that the password on a control command
--     is correct.
--
--  GIVEN          : 
--  GLOBAL USED    : 
--  RETURNS        : 
--
--  POST CONDITIONS: 
--
--  COMMENTS:
--      none
--
-----------------------------------------------------------------------*/
int
validate_password(str)
char   *str;
    {
    int   i, j;
    char  tstr[50];

    /* Find the start of the password. */

    for (j = 4, i = 0; j && i < strlen(str); i++) {
        if (str[i] == '.') j--;
        }

    /* If there was no 4th dot then there can't be a password. */

    if (j) return (0);

    /* Send it to the validation routine */

    strcpy(tstr, crypt(&str[i], "ck"));
    if (strcmp(&tstr[2], base_password) == 0) return (1);
    return (0);
    }
